# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from talisker import statsd


def test_parse_statsd_dsn_host():
    parsed = statsd.parse_statsd_dsn('udp://test.com')
    assert parsed == ('test.com', 8125, None, 512, False)


def test_parse_statsd_dsn_port():
    parsed = statsd.parse_statsd_dsn('udp://test.com:5000')
    assert parsed == ('test.com', 5000, None, 512, False)


def test_parse_statsd_dsn_prefix():
    parsed = statsd.parse_statsd_dsn('udp://test.com/foo.bar')
    assert parsed == ('test.com', 8125, 'foo.bar', 512, False)


def test_parse_statsd_dsn_prefix_with_slashes():
    parsed = statsd.parse_statsd_dsn('udp://test.com/foo/bar')
    assert parsed == ('test.com', 8125, 'foo.bar', 512, False)


def test_parse_statsd_dsn_size():
    parsed = statsd.parse_statsd_dsn('udp://test.com?maxudpsize=1024')
    assert parsed == ('test.com', 8125, None, 1024, False)


def test_parse_statsd_dsn_ipv6():
    parsed = statsd.parse_statsd_dsn('udp6://test.com')
    assert parsed == ('test.com', 8125, None, 512, True)



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import os
import sys
import collections
import functools
from ipaddress import ip_address, ip_network
from werkzeug.wrappers import Request, Response
from talisker import revision


class TestException(Exception):
    pass


NETWORKS = []
_loaded = False


def force_unicode(s):
    if isinstance(s, bytes):
        return s.decode('utf8')
    return s


def load_networks():
    networks = os.environ.get('TALISKER_NETWORKS', '').split(' ')
    return [ip_network(force_unicode(n)) for n in networks if n]


def private(f):
    """Only allow approved source addresses."""
    @functools.wraps(f)
    def wrapper(self, request):
        global NETWORKS, _loaded
        if not _loaded:
            NETWORKS = load_networks()
            _loaded = True
        if not request.access_route:
            # no client ip
            return Response(status='403')
        ip_str = request.access_route[0]
        if isinstance(ip_str, bytes):
            ip_str = ip_str.decode('utf8')
        ip = ip_address(ip_str)
        if ip.is_loopback or any(ip in network for network in NETWORKS):
            return f(self, request)
        else:
            return Response(status='403')
    return wrapper


class StandardEndpointMiddleware(object):
    """WSGI middleware to provide a standard set of endpoints for a service"""

    _ok_response = None

    @property
    def _ok(self):
        if self._ok_response is None:
            self._ok_response = Response(str(revision.get()))
        return self._ok_response

    def __init__(self, app, namespace='_status'):
        self.app = app
        self.namespace = namespace
        self.prefix = '/' + namespace

    def __call__(self, environ, start_response):
        request = Request(environ)
        if request.path.startswith(self.prefix):
            method = request.path[len(self.prefix):]
            if method == '':
                # no trailing /
                start_response('302', [('location', self.prefix + '/')])
                return ''
            if method == '/':
                method = 'index'
            else:
                method = method.lstrip('/')
            try:
                func = getattr(self, method)
            except AttributeError:
                response = Response(status=404)
            else:
                response = func(request)

            return response(environ, start_response)
        else:
            return self.app(environ, start_response)

    def index(self, request):
        methods = []
        item = '<li><a href="{0}"/>{0}</a> - {1}</li>'
        for name, func in list(self.__class__.__dict__.items()):
            if not name.startswith('_') and name != 'index':
                methods.append(item.format(name, func.__doc__))
        return Response(
            '<ul>' + '\n'.join(methods) + '<ul>', mimetype='text/html')

    def ping(self, request):
        """HAProxy status check"""
        return self._ok

    def check(self, request):
        """Nagios health check"""
        start = {}

        def nagios_start(status, headers, exc_info=None):
            # save status for inspection
            start['status'] = status
            start['headers'] = headers
            if exc_info:
                start['exc'] = sys.exc_info()

        response = self.app(request.environ, nagios_start)
        if not start:
            # nagios_start has not yet been called
            if isinstance(response, collections.Iterable):
                # force evaluation
                response = ''.join(response)

        if 'exc' in start:
            return Response('error', status=500)
        elif start.get('status', '').startswith('404'):
            # app does not provide /_status/nagios endpoint
            return self._ok
        else:
            # return app's response
            return Response(response,
                            status=start.get('status', 200),
                            headers=start.get('headers', []))

    @private
    def error(self, request):
        """Raise a TestError for testing"""
        raise TestException('this is a test, ignore')

    @private
    def metric(self, request):
        statsd = request.environ['statsd']
        statsd.incr('test')
        return Response('Incremented {}.test'.format(statsd._prefix))

    @private
    def info(self, request):
        return Response('Not Implemented', status=501)



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import os
from datetime import datetime
import logging

from gunicorn.instrument import statsd as gstatsd
from gunicorn.config import AccessLogFormat
from gunicorn.app.wsgiapp import WSGIApplication

from . import logs
from . import wsgi
from . import statsd


__all__ = ['access_log_format', 'logger_class']


class GunicornLogger(gstatsd.Statsd):
    """Custom gunicorn logger to use structured logging.

    Based on the statsd gunicorn logger, and also increases timestamp
    resolution to include msec in access and error logs.
    """

    # for access log
    def now(self):
        """return date in Apache Common Log Format, but with milliseconds"""
        formatted = datetime.utcnow().strftime('%d/%b/%Y:%H:%M:%S.%f')
        # trim to milliseconds, and hardcode TMZ, for standardising
        return '[' + formatted[:-3] + ' +0000]'

    def setup(self, cfg):
        super(GunicornLogger, self).setup(cfg)
        # gunicorn doesn't allow formatter customisation, so we need to alter
        # after setup
        error_handler = self._get_gunicorn_handler(self.error_log)
        error_handler.setFormatter(logs.StructuredFormatter())
        access_handler = self._get_gunicorn_handler(self.access_log)
        if access_handler:
            access_handler.setFormatter(
                logs.StructuredFormatter(self.access_fmt))

    @classmethod
    def install(cls):
        # in case used as a library, rather than via the entrypoint,
        # install the logger globally, as this is the earliest point we can do
        # so, if not using the talisker entry point
        logging.setLoggerClass(logs.StructuredLogger)


# gunicorn config
access_log_format = AccessLogFormat.default + ' duration=%(D)s'


class TaliskerApplication(WSGIApplication):
    def __init__(self, prog, devel=False):
        self._devel = devel
        super(TaliskerApplication, self).__init__(prog)

    def load_wsgiapp(self):
        app = super(TaliskerApplication, self).load_wsgiapp()
        app = wsgi.wrap(app)
        return app

    def init(self, parser, opts, args):
        """Provide talisker specific default config for gunicorn.

        Default config here can be overriden with cli args or config file."""

        cfg = super(TaliskerApplication, self).init(parser, opts, args)
        if cfg is None:
            cfg = {}

        cfg.update({
            'logger_class': GunicornLogger,
            'access_log_format': access_log_format,
            # level filtering controlled by handler, not logger
            'loglevel': 'notset'
        })

        # wire up statsd, if configured
        config = statsd.get_config()
        if 'hostport' in config and 'prefix' in config:
            cfg['statsd_host'] = config['hostport']
            cfg['statsd_prefix'] = config['prefix']

        # development config
        if self._devel:
            cfg['accesslog'] = '-'
            cfg['timeout'] = 99999

        return cfg


def parse_environ(environ):
    devel = 'DEVEL' in environ
    debug_log = environ.get('DEBUGLOG')
    return devel, debug_log


def run():  # pragma: no cover
    devel, debug = parse_environ(os.environ)
    logs.configure(devel, debug)
    app = TaliskerApplication(
        "%(prog)s [OPTIONS] [APP_MODULE]", devel)
    return app.run()


if __name__ == '__main__':
    run()  # pragma: no cover



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import os
import tempfile

import pytest
from werkzeug.test import Client
from werkzeug.wrappers import BaseResponse, Response, Request

import talisker.statsd
import talisker.endpoints
from talisker.endpoints import StandardEndpointMiddleware


@pytest.fixture
def wsgi_app(status='200', headers=[], body=''):
    def app(environ, start_response):
        start_response(status, headers)
        return body
    return app


@pytest.fixture
def client(wsgi_app):
    app = StandardEndpointMiddleware(wsgi_app)
    return Client(app, BaseResponse)


def set_networks(monkeypatch, networks):
    monkeypatch.setitem(os.environ, 'TALISKER_NETWORKS', networks)
    monkeypatch.setattr(talisker.endpoints, '_loaded', False)


@talisker.endpoints.private
def protected(self, request):
    return Response(status=200)


def test_private_no_config(monkeypatch):
    set_networks(monkeypatch, '')

    response = protected(None, Request({'REMOTE_ADDR': b'127.0.0.1'}))
    assert response.status_code == 200
    response = protected(None, Request({'REMOTE_ADDR': '127.0.0.1'}))
    assert response.status_code == 200
    response = protected(None, Request({'REMOTE_ADDR': b'1.2.3.4'}))
    assert response.status_code == 403
    response = protected(None, Request({'REMOTE_ADDR': '1.2.3.4'}))
    assert response.status_code == 403


def test_private_with_config(monkeypatch):
    set_networks(monkeypatch, '10.0.0.0/8')

    response = protected(None, Request({'REMOTE_ADDR': b'127.0.0.1'}))
    assert response.status_code == 200
    response = protected(None, Request({'REMOTE_ADDR': b'1.2.3.4'}))
    assert response.status_code == 403
    response = protected(None, Request({'REMOTE_ADDR': b'10.0.0.1'}))
    assert response.status_code == 200


def test_private_with_multiple_config(monkeypatch):
    set_networks(monkeypatch, '10.0.0.0/8 192.168.0.0/24')

    response = protected(None, Request({'REMOTE_ADDR': b'127.0.0.1'}))
    assert response.status_code == 200
    response = protected(None, Request({'REMOTE_ADDR': b'1.2.3.4'}))
    assert response.status_code == 403
    response = protected(None, Request({'REMOTE_ADDR': b'10.0.0.1'}))
    assert response.status_code == 200
    response = protected(None, Request({'REMOTE_ADDR': b'192.168.0.1'}))
    assert response.status_code == 200


def test_unknown_endpoint(client):
    response = client.get('/_status/unknown')
    assert response.status_code == 404


def test_pass_thru():
    c = client(wsgi_app(body='test'))
    response = c.get('/something')
    assert response.status_code == 200
    assert response.data == b'test'


def test_index_endpoint(client):
    response = client.get('/_status/')
    assert response.status_code == 200
    assert response.headers['Content-Type'] == 'text/html; charset=utf-8'


def test_index_redirect(client):
    response = client.get('/_status')
    assert response.status_code == 302
    assert response.headers['Location'] == '/_status/'


def test_ping(client, monkeypatch):
    monkeypatch.chdir(tempfile.mkdtemp())
    response = client.get('/_status/ping')
    assert response.status_code == 200
    assert response.headers['Content-Type'] == 'text/plain; charset=utf-8'
    assert response.data == b'unknown'


def test_check_no_app_url():
    c = client(wsgi_app('404'))
    response = c.get('/_status/check')
    assert response.status_code == 200
    assert response.headers['Content-Type'] == 'text/plain; charset=utf-8'
    assert response.data == b'unknown'


def test_check_with_app_url():

    def app(e, sr):
        """Implements custom check check"""
        if e['PATH_INFO'] == '/_status/check':
            sr('200', [])
            return b'app implemented check'
        else:
            sr('404', [])
            return ''

    c = client(app)
    response = c.get('/_status/check')
    assert response.data == b'app implemented check'


def test_check_with_no_app_url_iterator():

    def app(e, sr):
        yield 'app'
        sr('404', [])
        yield 'iterator'

    c = client(app)
    response = c.get('/_status/check')
    assert response.data == b'unknown'


def test_check_with_app_url_iterator():

    def app(e, sr):
        yield 'app'
        sr('200', [])
        yield 'iterator'

    c = client(app)
    response = c.get('/_status/check')
    assert response.data == b'appiterator'


def test_check_with_exc_info():
    def app(e, sr):
        try:
            raise Exception('test')
        except:
            sr(500, [], exc_info=1)
            return ''

    c = client(app)
    response = c.get('/_status/check')
    assert response.data == b'error'
    assert response.status_code == 500


def test_error(client):
    response = client.get('/_status/error',
                          environ_overrides={'REMOTE_ADDR': b'1.2.3.4'})
    assert response.status_code == 403
    with pytest.raises(talisker.endpoints.TestException):
        client.get('/_status/error',
                   environ_overrides={'REMOTE_ADDR': b'127.0.0.1'})


def test_metric(client):
    pipeline = talisker.statsd.get_client().pipeline()
    env = {'statsd': pipeline,
           'REMOTE_ADDR': b'127.0.0.1'}
    response = client.get('/_status/metric', environ_overrides=env)
    assert response.status_code == 200
    assert pipeline._stats[0] == 'test:1|c'



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

from urllib.parse import urlparse


def parse_url(url, proto='http'):
    # urlparse won't parse properly without a protocol
    if not url.startswith(proto + '://'):
        url = proto + '://' + url
    return urlparse(url)



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys
from gunicorn.config import Config

from talisker import gunicorn
from talisker import logs
from talisker import statsd

from freezegun import freeze_time


@freeze_time('2016-01-02 03:04:05.6789')
def test_gunicorn_logger_now():
    logger = gunicorn.GunicornLogger(Config())
    ts = logger.now()
    assert ts == '[02/Jan/2016:03:04:05.678 +0000]'


def test_gunicorn_logger_set_formatters_on_gunicorn_logs():
    cfg = Config()
    cfg.set('accesslog', '-')
    logger = gunicorn.GunicornLogger(cfg)
    error = logger._get_gunicorn_handler(logger.error_log)
    assert isinstance(error.formatter, logs.StructuredFormatter)
    access = logger._get_gunicorn_handler(logger.access_log)
    assert isinstance(access.formatter, logs.StructuredFormatter)


def test_gunicorn_application_init(monkeypatch):
    monkeypatch.setattr(sys, 'argv', ['talisker', 'wsgi:app'])
    app = gunicorn.TaliskerApplication('')
    assert app.cfg.access_log_format == gunicorn.access_log_format
    assert app.cfg.logger_class == gunicorn.GunicornLogger


def test_gunicorn_application_init_statsd(monkeypatch):
    monkeypatch.setattr(sys, 'argv', ['talisker', 'wsgi:app'])
    statsd_config = {'hostport': '1.2.3.4:2000', 'prefix': 'test'}
    monkeypatch.setattr(statsd, 'get_config', lambda: statsd_config)
    app = gunicorn.TaliskerApplication('')
    assert app.cfg.statsd_host == ('1.2.3.4', 2000)
    assert app.cfg.statsd_prefix == 'test'


def test_gunicorn_application_init_devel(monkeypatch):
    monkeypatch.setattr(sys, 'argv', ['talisker', 'wsgi:app'])
    app = gunicorn.TaliskerApplication('', devel=True)
    assert app.cfg.accesslog == '-'
    assert app.cfg.timeout == 99999


def test_gunicorn_application_init_devel_overriden(monkeypatch):
    monkeypatch.setattr(
        sys, 'argv',
        ['talisker', 'wsgi:app', '--timeout', '10'])
    app = gunicorn.TaliskerApplication('', devel=True)
    assert app.cfg.timeout == 10


def wsgi(environ, start_response):
    start_response('200 OK', [])
    return ''


def test_gunicorn_application_load(monkeypatch):
    monkeypatch.setattr(
        sys, 'argv', ['', __name__ + ':wsgi', '--bind=0.0.0.0:0'])
    app = gunicorn.TaliskerApplication('')
    wsgiapp = app.load_wsgiapp()
    assert wsgiapp._talisker_wrapped
    assert wsgiapp._talisker_original_app == wsgi


def test_parse_environ():
    parse = gunicorn.parse_environ
    assert parse({}) == (False, None)
    assert parse({'DEVEL': 1}) == (True, None)
    assert parse({'DEBUGLOG': '/tmp/log'}) == (False, '/tmp/log')
    assert parse({'DEVEL': 1, 'DEBUGLOG': '/tmp/log'}) == (True, '/tmp/log')



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import pprint


def application(environ, start_response):
    status = '404 Not Found'
    start_response(status, [('content-type', 'text/plain')])
    s = environ['statsd']
    statsd = 'statsd: {} {}'.format(s._addr, s._prefix)
    output = (status + '\n\n' +
              statsd + '\n\n' +
              pprint.pformat(environ))

    return output.encode('utf8')



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from datetime import timedelta
import requests
import talisker.requests


# @pytest.fixture
def response(
        method='GET',
        host='http://example.com',
        url='/',
        code=200,
        elapsed=1.0):
    req = requests.Request(method, host + url)
    resp = requests.Response()
    resp.request = req.prepare()
    resp.status_code = code
    resp.elapsed = timedelta(seconds=elapsed)
    return resp


def test_metric_hook_root():
    r = response()
    name, duration = talisker.requests.get_timing(r)
    assert name == 'requests.example-com.GET.200'
    assert duration == 1000.0


def test_metric_hook_post():
    r = response(method='POST')
    name, duration = talisker.requests.get_timing(r)
    assert name == 'requests.example-com.POST.200'
    assert duration == 1000.0


def test_metric_hook_500():
    r = response(code=500)
    name, duration = talisker.requests.get_timing(r)
    assert name == 'requests.example-com.GET.500'
    assert duration == 1000.0



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

from functools import wraps
import uuid

from .request_context import request_context, cleanup
from .logs import set_logging_context

HEADER = 'X-Request-Id'


def generate():
    return str(uuid.uuid4()).encode('utf8')


def get():
    try:
        return request_context.request_id
    except AttributeError:
        return ""


def set(id):
    """Sets id in both general request context, and specific logging dict"""
    request_context.request_id = id
    set_logging_context(request_id=id)


def decorator(id_func):
    """Decorator to set a thread local request id for function.

    Takes a function that will return the request id from the function params,
    so you can configure it. Cleans up on the way out.
    """
    def wrapper(func):
        @wraps(func)
        def decorator(*args, **kwargs):
            id = id_func(*args, **kwargs)
            set(id)
            try:
                return func(*args, **kwargs)
            finally:
                cleanup()
        return decorator
    return wrapper


class RequestIdMiddleware(object):
    """WSGI middleware to set the request id."""

    def __init__(self, app, header=HEADER):
        self.app = app
        self.header = header.encode('utf8')
        self.wsgi_header = 'HTTP_' + header.upper().replace('-', '_')
        self.wsgi_header = self.wsgi_header.encode('utf8')

    def __call__(self, environ, start_response):
        if self.wsgi_header not in environ:
            environ[self.wsgi_header] = generate()
        id = environ[self.wsgi_header]
        set(id)
        environ[b'REQUEST_ID'] = id
        return self.app(environ, start_response)



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

from . import request_id
from . import request_context
from . import endpoints
from . import statsd
from . import requests
from . import revision


def set_environ(app, **kwargs):
    def middleware(environ, start_response):
        for key, value in kwargs.items():
            environ[key] = value
        return app(environ, start_response)
    return middleware


def set_headers(app, add_headers):
    def middleware(environ, start_response):
        def custom_start_response(status, headers, exc_info=None):
            for header, value in add_headers.items():
                headers.append((header, value))
            return start_response(status, headers, exc_info)
        return app(environ, custom_start_response)
    return middleware


def wrap(app):
    if getattr(app, '_talisker_wrapped', False):
        return app

    wrapped = app
    # added in reverse order
    # expose some standard endpoint
    wrapped = set_headers(wrapped, {'X-VCS-Revision': revision.header()})
    wrapped = endpoints.StandardEndpointMiddleware(wrapped)
    # set some standard environ items
    wrapped = set_environ(
        wrapped,
        statsd=statsd.get_client(),
        requests=requests.get_session(),
    )
    # add request id info to thread locals
    wrapped = request_id.RequestIdMiddleware(wrapped)
    # clean up request context on the way out
    wrapped = request_context.cleanup_middleware(wrapped)
    wrapped._talisker_wrapped = True
    wrapped._talisker_original_app = app
    return wrapped



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

from collections import OrderedDict
from contextlib import contextmanager
import time
import logging
import sys

from .request_context import request_context


_logging_configured = False


NOISY_LOGS = {
    'requests': logging.WARNING,
}


def set_logging_context(extra=None, **kwargs):
    """Update structured logging keys in a thread/context dict."""
    if not hasattr(request_context, 'extra'):
        request_context.extra = {}
    if extra:
        request_context.extra.update(extra)
    if kwargs:
        request_context.extra.update(kwargs)


@contextmanager
def extra_logging(extra=None, **kwargs):
    set_logging_context(extra, **kwargs)
    yield
    remove = list(kwargs.items())
    if extra:
        remove.extend(extra.items())
    for k, v in remove:
        request_context.extra.pop(k, None)


def add_talisker_handler(root, level, handler):
    handler.setFormatter(StructuredFormatter())
    handler.setLevel(level)
    root.addHandler(handler)


def configure(devel=False, debug=None, tags=None):
    """Configure default logging setup for our services.

    This is basically:
     - log to stderr
     - output hybrid logfmt structured format
     - add some basic structured data by default
    """

    # avoid duplicate logging
    global _logging_configured
    if _logging_configured:
        return

    logging.setLoggerClass(StructuredLogger)
    if tags is not None:
        StructuredLogger.update_extra(tags)

    root = logging.getLogger()
    root.setLevel(logging.NOTSET)
    add_talisker_handler(root, logging.INFO, logging.StreamHandler())
    configure_warnings(devel)
    supress_noisy_logs()

    # defer this until logging has been set up
    logger = logging.getLogger(__name__)

    if debug is not None:
        if can_write_to_file(debug):
            handler = logging.handlers.TimedRotatingFileHandler(
                debug,
                when='D',
                interval=1,
                backupCount=0,
                delay=True,
                utc=True,
            )
            add_talisker_handler(root, logging.DEBUG, handler)
            logger.info('enabling debug logs', extra={'path': debug})
        else:
            logger.info('could not enable debug log, could not write to path',
                        extra={'path': debug})

    _logging_configured = True


def can_write_to_file(path):
    try:
        open(path, 'a').close()
    except:
        return False
    else:
        return True


def supress_noisy_logs():
    """Set some custom log levels on some sub logs"""
    for name, level in NOISY_LOGS.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)


def configure_warnings(enable):
    # never propogate warnings to root
    warnings = logging.getLogger('py.warnings')
    warnings.propagate = False

    if enable:
        warnings.addHandler(logging.StreamHandler())


def configure_test_logging():
    """Add a NullHandler to root logger.

    Prevents unconfigured logging from erroring, and swallows all logging,
    which is usually what you want for unit tests.  Unit test fixtures can
    still add their own loggers to assert against log messages if needed.
    """

    handler = logging.NullHandler()
    add_talisker_handler(logging.getLogger(), logging.NOTSET, handler)
    configure_warnings(True)


class StructuredLogger(logging.Logger):
    """A logger that handles passing 'extra' arguments to all logging calls.

    Supports 3 sources of extra structured data:

    1) global extra, designed to be set once at process start/
    2) context extra, designed to be set per request or job, can cleaned up
       afterwards.
    3) per call extra, passed by the log call, as per normal logging
       e.g. log.info('...', extra={...})
       These keys are prefixed by cls._prefix to avoid collisions.

    """

    _extra = OrderedDict()
    _prefix = 'svc.'
    structured = True

    @classmethod
    def update_extra(cls, extra):
        cls._extra.update(extra)

    # sadly, we must subclass and override, rather that use the new
    # setLogRecordFactory() in 3.2+, as that does not pass the extra args
    # through. Also, we need to support python 2.
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        # at this point we have 3 possible sources of extra kwargs
        # - log call: extra
        # - context : local.extra
        # - global  : cls._extra
        # These are added in order of most specific to least, to push line
        # noise to the end of the log line
        all_extra = OrderedDict()
        if extra is not None:
            # prefix call site extra args, to avoid collisions
            for k, v in list(extra.items()):
                all_extra[self._prefix + k] = v
        all_extra.update(getattr(request_context, 'extra', {}))
        all_extra.update(
            (k, v) for k, v in self._extra.items() if k not in all_extra)
        kwargs = dict(func=func, extra=all_extra, sinfo=sinfo)
        # python 2 doesn't support sinfo parameter
        if sys.version_info[0] == 2:
            kwargs.pop('sinfo')
        record = super(StructuredLogger, self).makeRecord(
            name, level, fn, lno, msg, args, exc_info, **kwargs)
        # store extra explicitly for StructuredFormatter to use
        if self.structured:
            record._structured = all_extra
        return record


class StructuredFormatter(logging.Formatter):
    """Add additional structured data in logfmt style to formatted log.

    Outputs the specified format as normal, and adds any structured data
    available in logfmt. Requires StructuredLogger to work.

    e.g.

    2016-01-13 10:24:07.357Z INFO name "my message" foo=data bar="other data"

    """

    FORMAT = '%(asctime)s.%(msecs)03dZ %(levelname)s %(name)s "%(message)s"'
    DATEFMT = "%Y-%m-%d %H:%M:%S"

    # use utc time. No idea why this is not the default.
    converter = time.gmtime

    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = self.FORMAT
        if datefmt is None:
            datefmt = self.DATEFMT
        super(StructuredFormatter, self).__init__(fmt, datefmt)

    def getMessage(self):
        msg = super(self, StructuredFormatter).getMessage()
        return self.escape_quotes(msg)

    def format(self, record):
        """Render the format, adding any extra as structured tags."""
        # this is verbatim from the parent class in stdlib
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self._fmt % record.__dict__

        # add our structured tags *before* exception info is added
        structured = getattr(record, '_structured', {})
        if structured:
            logfmt = (self.logfmt(k, record.__dict__[k]) for k in structured)
            s += " " + " ".join(logfmt)

        # this is verbatim from the parent class in stdlib
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            try:
                s = s + record.exc_text
            except UnicodeError:
                # Sometimes filenames have non-ASCII chars, which can lead
                # to errors when s is Unicode and record.exc_text is str
                # See issue 8924.
                # We also use replace for when there are multiple
                # encodings, e.g. UTF-8 for the filesystem and latin-1
                # for a script. See issue 13232.
                s = s + record.exc_text.decode(sys.getfilesystemencoding(),
                                               'replace')
        return s

    def escape_quotes(self, s):
        return s.replace('"', '\\"')

    def remove_quotes(self, s):
        return s.replace('"', '')

    def logfmt(self, k, v):
        # we need unicode strings so as to be able to replace
        if isinstance(k, bytes):
            k = k.decode('utf8')
        if not isinstance(v, str):
            if isinstance(v, bytes):
                v = v.decode('utf8')
            elif not isinstance(v, str):
                v = str(v)

        k = k.replace(' ', '_').replace('"', '')
        k = self.remove_quotes(k)
        v = self.remove_quotes(v)
        if ' ' in v:
            v = '"' + v + '"'
        return "%s=%s" % (k, v)



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import uuid
from wsgiref.util import setup_testing_defaults

from py.test import fixture
import mock

from .fixtures import clean_up_context  # noqa
from talisker import request_id
from talisker.request_context import request_context


@fixture
def environ():
    env = {}
    setup_testing_defaults(env)
    return env


@fixture
def start_response():
    return lambda: None


@fixture
def id():
    return str(uuid.uuid4()).encode('utf8')


def make_assert_app(id):
    def app(environ, start_response):
        assert environ[b'REQUEST_ID'] == id
        assert request_id.get() == id
        assert request_context.extra['request_id'] == id
    return app


def test_middleware_with_id(environ, start_response, id):
    environ[b'HTTP_X_REQUEST_ID'] = id
    app = make_assert_app(id)
    middleware = request_id.RequestIdMiddleware(app)
    middleware(environ, start_response)


def test_middleware_without_id(environ, start_response, id):
    app = make_assert_app(id)
    middleware = request_id.RequestIdMiddleware(app)
    with mock.patch('talisker.request_id.generate') as mock_id:
        mock_id.return_value = id
        middleware(environ, start_response)


def test_middleware_alt_header(environ, start_response, id):
    environ[b'HTTP_X_ALTERNATE'] = id
    app = make_assert_app(id)
    middleware = request_id.RequestIdMiddleware(app, 'X-Alternate')
    middleware(environ, start_response)


def test_decorator(id):

    @request_id.decorator(lambda: id)
    def test():
        assert request_id.get() == id

    test()



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

from subprocess import check_output as run
import sys
import tempfile

from talisker import revision
from pytest import mark


def is_git_configured():
    try:
        git = run(['which', 'git']).strip()
    except:
        return False
    return git != ""


def is_bzr_configured():
    try:
        bzr = run(['which', 'bzr']).strip()
    except:
        return False
    return bzr != ""


requires_bzr = mark.skipif(
    not is_bzr_configured(),
    reason='bzr not installed/configured')

requires_git = mark.skipif(
    not is_git_configured(),
    reason='git not installed/configured')


@requires_git
def test_git(monkeypatch):
    dir = tempfile.mkdtemp()
    monkeypatch.chdir(dir)
    run(['git', 'init', '.'])
    run(['git', 'config', 'user.email', 'someone@email.com'])
    run(['git', 'config', 'user.name', 'someone'])
    run(['touch', 'foo'])
    run(['git', 'add', 'foo'])
    run(['git', 'commit', '-m', 'init'])
    rev = revision.git()
    assert len(rev.strip()) == 40


def set_up_bzr():
    run(['bzr', 'init', '.'])
    run(['bzr', 'whoami', 'someone@email.com', '--branch'])
    run(['touch', 'foo'])
    run(['bzr', 'add', 'foo'])
    run(['bzr', 'commit', '-m', 'init'])


@requires_bzr
def test_bzr(monkeypatch):
    dir = tempfile.mkdtemp()
    monkeypatch.chdir(dir)
    set_up_bzr()
    rev = revision.bzr()
    assert rev.strip() == b'1'


@requires_bzr
@mark.skipif(sys.version_info >= (3, 0), reason="requires python2")
def test_bzr_version_info_py2(monkeypatch):
    dir = tempfile.mkdtemp()
    monkeypatch.chdir(dir)
    monkeypatch.syspath_prepend(dir)
    set_up_bzr()
    vinfo = run(['bzr', 'version-info', '--format=python'])
    with open('versioninfo.py', 'wb') as f:
        f.write(vinfo)
    rev = revision.bzr_version_info()
    assert rev.strip() == '1'



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import logging
import logging.handlers


import shlex
import time

from py.test import fixture

from .fixtures import clean_up_context  # noqa
from talisker import logs
from talisker.request_context import request_context

TIME = time.mktime((2016, 1, 17, 12, 30, 10, 1, 48, 0))
MSEC = 123.456
TIMESTAMP = "2016-01-17 12:30:10.123Z"


@fixture
def record_args():
    """Test arguments to the makeRecord function."""
    return ('name', logging.INFO, 'fn', 'lno', 'msg here', tuple(), None)


def make_record(extra):
    """Make a test record from StructuredLogger."""
    logger = logs.StructuredLogger('test')
    record = logger.makeRecord(*record_args(), extra=extra)
    # stub out the time
    record.__dict__['created'] = TIME
    record.msecs = MSEC
    return record


def parse_logfmt(log):
    """Stupid simple logfmt parser"""
    parsed = shlex.split(log)
    date, time, level, name, msg = parsed[:5]
    extra = dict((v.split('=')) for v in parsed[5:])
    return date + " " + time, level, name, msg, extra


def test_set_logging_context_no_extra():
    if hasattr(request_context, 'extra'):
        del request_context.extra
    logs.set_logging_context()
    assert request_context.extra == {}


def test_set_logging_context():
    logs.set_logging_context(a=1)
    assert request_context.extra == {'a': 1}


def test_make_record_no_extra(record_args):
    logger = logs.StructuredLogger('test')
    record = logger.makeRecord(*record_args)
    assert record._structured == {}


def test_make_record_global_extra(record_args):
    logger = logs.StructuredLogger('test')
    logger.update_extra({'a': 1})
    record = logger.makeRecord(*record_args)
    assert record.__dict__['a'] == 1
    assert record._structured == {'a': 1}


def test_make_record_context_extra(record_args):
    logger = logs.StructuredLogger('test')
    logs.set_logging_context(a=1)
    record = logger.makeRecord(*record_args)
    assert record.__dict__['a'] == 1
    assert record._structured == {'a': 1}


def test_make_record_all_extra(record_args):
    logger = logs.StructuredLogger('test')
    logger.update_extra({'a': 1})
    logs.set_logging_context(b=2)
    record = logger.makeRecord(*record_args, extra={'c': 3})

    assert record.__dict__['a'] == 1
    assert record.__dict__['b'] == 2
    assert record.__dict__['c'] == 3
    assert record._structured == {'a': 1, 'b': 2, 'c': 3}


def test_make_record_context_overiddes(record_args):
    logger = logs.StructuredLogger('test')
    logger.update_extra({'a': 1})
    logs.set_logging_context(a=2)
    record = logger.makeRecord(*record_args)

    assert record.__dict__['a'] == 2
    assert record._structured == {'a': 2}


def test_formatter_no_args():
    fmt = logs.StructuredFormatter()
    log = fmt.format(make_record({}))
    timestamp, level, name, msg, structured = parse_logfmt(log)
    assert timestamp == TIMESTAMP
    assert level == 'INFO'
    assert name == 'name'
    assert msg == "msg here"
    assert structured == {}


def test_formatter_with_extra():
    fmt = logs.StructuredFormatter()
    log = fmt.format(make_record({'foo': 'bar', 'baz': 'with spaces'}))
    timestamp, level, name, msg, structured = parse_logfmt(log)
    assert timestamp == TIMESTAMP
    assert level == 'INFO'
    assert name == 'name'
    assert msg == "msg here"
    assert structured == {
        'foo': 'bar',
        'baz': 'with spaces',
    }


def test_formatter_with_exec_info():
    fmt = logs.StructuredFormatter()
    record = make_record({'foo': 'bar'})
    record.exc_info = True
    record.exc_text = "Traceback:\none\ntwo\nthree"
    log = fmt.format(record)
    lines = log.splitlines()
    timestamp, level, name, msg, structured = parse_logfmt(lines[0])
    assert structured['foo'] == 'bar'
    assert lines[1] == 'Traceback:'
    assert lines[2] == 'one'
    assert lines[3] == 'two'
    assert lines[4] == 'three'


def test_configure(capsys):
    logs.configure(logging.INFO, tags=dict(foo='bar baz'))
    logger = logging.getLogger('test')
    logger.info('test msg')
    out, err = capsys.readouterr()
    assert out == ""
    assert err
    timestamp, level, name, msg, structured = parse_logfmt(err)
    assert level == 'INFO'
    assert name == 'test'
    assert msg == 'test msg'
    assert structured['foo'] == 'bar baz'


def test_escape_quotes():
    fmt = logs.StructuredFormatter()
    assert fmt.escape_quotes('foo') == 'foo'
    assert fmt.escape_quotes('foo "bar"') == r'foo \"bar\"'


def test_logfmt():
    fmt = logs.StructuredFormatter()
    assert fmt.logfmt('foo', 'bar') == 'foo=bar'
    assert fmt.logfmt('foo', 'bar baz') == 'foo="bar baz"'
    assert fmt.logfmt('foo', '"baz"') == r'foo=baz'
    assert fmt.logfmt('foo', 'bar "baz"') == r'foo="bar baz"'
    assert fmt.logfmt('foo', b'bar') == r'foo=bar'
    assert fmt.logfmt(b'foo', 'bar') == r'foo=bar'
    assert fmt.logfmt('foo foo', 'bar') == r'foo_foo=bar'
    assert fmt.logfmt('foo"', 'bar') == r'foo=bar'
    assert fmt.logfmt('foo"', 1) == r'foo=1'



# Copyright (C) 2016- Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import *  # noqa

import os
from urllib.parse import urlparse, parse_qs
from statsd import StatsClient, defaults


_client = None


class TaliskerStatsdClient(StatsClient):
    """Statsd client that exposes it's config for inspection"""
    def __init__(self,
                 host=defaults.HOST,
                 port=defaults.PORT,
                 prefix=defaults.PREFIX,
                 maxudpsize=defaults.MAXUDPSIZE,
                 ipv6=defaults.IPV6):

        # store ipv6 config or else it's buried in socket
        self.ipv6 = ipv6

        super(TaliskerStatsdClient, self).__init__(
            host, port, prefix, maxudpsize, ipv6)

    @property
    def port(self):
        return self._addr[1]

    @property
    def prefix(self):
        return self._prefix

    @property
    def hostport(self):
        return self.hostname + ':' + str(self.port)

    @property
    def maxudpsize(self):
        return self._maxudpsize


def get_config():
    client = get_client()
    if isinstance(client, DummyClient):
        return {}
    else:
        return {
            'host': client.hostname,
            'hostport': client.hostport,
            'port': client.port,
            'prefix': client.prefix,
            'maxudpsize': client.maxudpsize,
            'ipv6': client.ipv6,
        }


def parse_statsd_dsn(dsn):
    parsed = urlparse(dsn)
    host = parsed.hostname
    port = parsed.port or defaults.PORT
    prefix = None
    if parsed.path:
        prefix = parsed.path.strip('/').replace('/', '.')
    ipv6 = parsed.scheme in ('udp6', 'tcp6')
    size = int(
        parse_qs(parsed.query).get('maxudpsize', [defaults.MAXUDPSIZE])[0])
    return host, port, prefix, size, ipv6


def get_client(dsn=None):
    global _client

    if _client is None:
        if dsn is None:
            dsn = os.environ.get('STATSD_DSN', None)
        if dsn is None:
            _client = DummyClient()
        else:
            if not dsn.startswith('udp'):
                raise Exception('Talisker only supports udp stastd client')
            _client = TaliskerStatsdClient(*parse_statsd_dsn(dsn))

    return _client


class DummyClient(TaliskerStatsdClient):

    def _after(self, stat):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

install_requires = [
    'gunicorn>=19.5.0',
    'Werkzeug>=0.11.5',
    'statsd>=3.2.1',
    'requests>=2.10.0',
    'future>=0.15.2',
]

if sys.version_info < (3, 3):
    install_requires.append('ipaddress==1.0.16')

setup(
    name='talisker',
    version='0.5.4',
    description="A common WSGI stack",
    long_description=readme + '\n\n' + history,
    author="Simon Davy",
    author_email='simon.davy@canonical.com',
    url='https://github.com/canonical-ols/talisker',
    packages=['talisker'],
    package_dir={'talisker': 'talisker'},
    package_data={'talisker': ['logstash/*']},
    include_package_data=True,
    license="GPL3",
    zip_safe=False,
    keywords='talisker',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: System :: Logging',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=install_requires,
    test_suite='tests',
    setup_requires=['pytest-runner'],
    entry_points={
        'console_scripts': ['talisker=talisker.gunicorn:run'],
    }
)
